package ca.recrown.islandsurvivalcraft.world.Information;

import java.lang.Thread.State;
import java.util.HashSet;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import ca.recrown.islandsurvivalcraft.utilities.GeneralUtilities;
import ca.recrown.islandsurvivalcraft.utilities.caching.Cache;
import ca.recrown.islandsurvivalcraft.utilities.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMap;

public class IslandInformationManager implements Runnable {
    private final IslandWorldMap islandMap;
    private final Thread loader;
    private final Cache<Point2, CompletableFuture<HashSet<IslandInformation>>> buildingCache = new Cache<>();
    private final ConcurrentHashMap<Point2, HashSet<IslandInformation>> islandSets = new ConcurrentHashMap<>();

    public IslandInformationManager(IslandWorldMap islandWorldMap) {
        this.islandMap = islandWorldMap;
        loader = new Thread(this);
        loader.setDaemon(true);
        loader.setName("Island-Info-Builder");
        loader.setPriority(Thread.MIN_PRIORITY);
        loader.start();
    }

    private HashSet<IslandInformation> buildChunkIslandInformation(Point2 chunkCoords) {
        HashSet<IslandInformation> result = new HashSet<>();
        Point2[] surroundings = new Point2[4];
        surroundings[0] = chunkCoords.shift(1, 0);
        surroundings[1] = chunkCoords.shift(-1, 0);
        surroundings[2] = chunkCoords.shift(0, -1);
        surroundings[3] = chunkCoords.shift(0, 1);
        HashSet<Point2> checkedCoordinates = new HashSet<>();
        for (Point2 surrounding : surroundings) {
            if (Thread.currentThread().isInterrupted())
                return null;
            HashSet<IslandInformation> chunkIslandSet = islandSets.get(surrounding);
            if (chunkIslandSet != null) {
                for (IslandInformation islandInformation : chunkIslandSet) {
                    if (islandInformation.isIslandInChunk(chunkCoords)) {
                        checkedCoordinates.addAll(islandInformation.getIslandCoordinates());
                        result.add(islandInformation);
                    }
                }
            }
        }

        for (int x = 0; x < GeneralUtilities.CHUNK_SIZE; x++) {
            for (int z = 0; z < GeneralUtilities.CHUNK_SIZE; z++) {
                if (Thread.currentThread().isInterrupted())
                    return null;
                int worldX = chunkCoords.x * GeneralUtilities.CHUNK_SIZE + x;
                int worldZ = chunkCoords.y * GeneralUtilities.CHUNK_SIZE + z;
                if (islandMap.isIsland(worldX, worldZ)) {
                    IslandInformationBuilder infoBuilder = new IslandInformationBuilder();
                    Point2 origin = islandMap.findIslandOrigin(worldX, worldZ, (p) -> {
                        if (Thread.currentThread().isInterrupted())
                            return true;
                        if (!checkedCoordinates.add(p))
                            return true;

                        HashSet<IslandInformation> chunkIslandSet = islandSets.get(GeneralUtilities.worldToChunkCoordinates(p));
                        if (chunkIslandSet != null) {
                            for (IslandInformation islandInformation : chunkIslandSet) {
                                if (islandInformation.isIslandInChunk(chunkCoords)) {
                                    checkedCoordinates.addAll(islandInformation.getIslandCoordinates());
                                    result.add(islandInformation);
                                    return true;
                                }
                            }
                        }
                        infoBuilder.addCoordinate(p, islandMap.isEdgeOfIsland(p.x, p.y));
                        infoBuilder.addChunk(GeneralUtilities.worldToChunkCoordinates(p));
                        return false;
                    });
                    if (origin != null) {
                        IslandInformation islandInfo = infoBuilder.build(origin);
                        result.add(islandInfo);
                    }
                }
            }
        }
        return result;
    }

    public void loadChunkIslandInformation(Point2 chunkCoords) {
        CompletableFuture<HashSet<IslandInformation>> completableFuture = new CompletableFuture<>();
        buildingCache.set(chunkCoords, completableFuture);
        if (loader.getState() == State.WAITING) {
            synchronized (loader) {
                loader.notifyAll();
            }
        }
    }

    public void unloadChunkIslandInformation(Point2 chunkCoords) {
        if (buildingCache.contains(chunkCoords))
            buildingCache.remove(chunkCoords).cancel(true);
        islandSets.remove(chunkCoords);
    }

    /**
     * Gets the island information for the given coordinates.
     * 
     * @param coords The coordinates to check for island information.
     * @param checkLoading Whether or not to check for the currently loading chunks.
     * @return The island information or null if there is no island at those
     *         coordinates.
     */
    public IslandInformation getIslandInformation(Point2 coords, boolean checkLoading) {
        Point2 chunkCoords = GeneralUtilities.worldToChunkCoordinates(coords);
        if (!islandSets.containsKey(chunkCoords)) {
            if (checkLoading && buildingCache.contains(chunkCoords)) {
                try {
                    buildingCache.get(chunkCoords).get();
                } catch (InterruptedException | ExecutionException e) {
                }
            }
        }
        HashSet<IslandInformation> chunkIslands = islandSets.get(chunkCoords);

        if (chunkIslands != null) {
            for (IslandInformation islandInformation : chunkIslands) {
                if (islandInformation.isWithinIsland(coords))
                    return islandInformation;
            }
        }
        return null;
    }

    @Override
    public void run() {
        while (!Thread.currentThread().isInterrupted()) {
            while (buildingCache.size() > 0) {
                Point2 chunkCoords = buildingCache.getMostRecentKey();
                CompletableFuture<HashSet<IslandInformation>> future = buildingCache.remove(chunkCoords);
                HashSet<IslandInformation> islandInformations = buildChunkIslandInformation(chunkCoords);
                islandSets.put(chunkCoords, islandInformations);
                future.complete(islandInformations);
            }
            try {
                synchronized (loader) {
                    loader.wait();
                }
            } catch (InterruptedException e) {
            }
        }
    }
}